greeks = ('alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta',
          'eta', 'theta', 'iota', 'kappa', 'lambda', 'mu', 'nu',
          'xi', 'omicron', 'pi', 'rho', 'sigma', 'tau', 'upsilon',
          'phi', 'chi', 'psi', 'omega')
